# coding=utf-8
import logging
import logging.handlers
import threading
import time

import PyTango

# Definition of errors
ERR_KEITHLEY_COMM_TIMEOUT = - 77
ERR_KEITHLEY_BAD_MODE = - 78
ERR_KEITHLEY_NO_PARAMS = - 79

ERR_KEITHLEY_INVALID_STATUS = - 81
ERR_KEITHLEY_BAD_PARAMS = - 82
ERR_KEITHLEY_BAD_CHANNEL = - 83
ERR_KEITHLEY_GPIB_WRITE_READ = - 84
ERR_KEITHLEY_GPIB_READ = -  85
ERR_KEITHLEY_GPIB_WRITE = - 86
ERR_KEITHLEY_BAD_MODEL = - 87
ERR_KEITHLEY_INIT_IGNORED = - 88

ERR_KEITHLEY_WRONG_PARAMS = - 90
ERR_KEITHLEY_WRONG_PARAMS_FUNCTION = - 91
ERR_KEITHLEY_WRONG_PARAMS_OUTPUT_TYPE = - 92
ERR_KEITHLEY_WRONG_PARAMS_OUTPUT_ELEMENTS = - 93
ERR_KEITHLEY_WRONG_PARAMS_POWERON = - 94
ERR_KEITHLEY_WRONG_PARAMS_AUTOZERO = - 95
ERR_KEITHLEY_WRONG_PARAMS_AMETHOD = - 96
ERR_KEITHLEY_WRONG_PARAMS_INPUT_SOURCE = - 97
ERR_KEITHLEY_WRONG_PARAMS_COUPLE_TRIGG = - 98



class KeithleyError(Exception):
    """
        Class to define and return the possible errors of the Agient PS
    """

    def __init__(self, value, *args):
        """
        Method to acquire the error code value and possible args
        :param value: input error code
        :param args: other possible params.
        """
        self.errcode = value
        self.args = args

    def __str__(self):
        """
        Method to retrieve the error message to output
        :return message: Error code message
        """
        return self._get_error_msg(self.errcode, self.args)

    def _get_error_msg(self, value, var):
        """
        Method to parse the error code and retrieve its associated message.
        Done as if else structure and not in dict structure in behalf of read
        and edit easiness.
        :return msg: Error code message
        """
        if value == '01':
            msg = 'Error %s: Unknown command' % value
        elif value == '02':
            msg = 'Error %s: Unknown argument' % value
        elif value == '03':
            msg = 'Error %s: Index out of range' % value
        elif value == '04':
            msg = 'Error %s: Not Enough Arguments' % value
        elif value == '05':
            msg = 'Error %s: Privilege Level Requirement not met' % value
        elif value == '06':
            msg = 'Error %s: Saving Error on device' % value
        elif value == '07':
            msg = 'Error %s: Invalid password' % value
        elif value == '08':
            msg = 'Error %s: Power supply in fault' % value
        elif value == '09':
            msg = 'Error %s: Power supply already ON' % value
        elif value == '10':
            msg = 'Error %s: Setpoint is out of model limits' % value
        elif value == '11':
            msg = 'Error %s: Setpoint is out of software limits' % value
        elif value == '12':
            msg = 'Error %s: Setpoint is not a number' % value
        elif value == '13':
            msg = 'Error %s: Module is OFF' % value
        elif value == '14':
            msg = 'Error %s: Slew Rate out of limits' % value
        elif value == '15':
            msg = 'Error %s: Device is set in local mode' % value
        elif value == '16':
            msg = 'Error %s: Module is not in waveform mode' % value
        elif value == '17':
            msg = 'Error %s:Module is in waveform mode' % value
        elif value == '18':
            msg = 'Error %s: Device is set in remote mode' % value
        elif value == '19':
            msg = ('Error %s: Module is already in the selected loop mode'
                   % value)
        elif value == '20':
            msg = 'Error %s: Module is not in the selected loop mode' % value
        elif value == ERR_KEITHLEY_COMM_TIMEOUT:
            msg = ('Error %s: Error timed out, the device did not respond. '
                   % value)
        elif value == ERR_KEITHLEY_BAD_MODE:
            msg = ('Error %s: Error of loop mode, the device is not set to the'
                   ' correct mode. ' % value)
        elif value == ERR_KEITHLEY_NO_PARAMS:
            msg = ('Error %s: Error of parameters. Parameter expected, none '
                   'given' % value)
        elif value == ERR_KEITHLEY_WRONG_PARAMS:
            msg = ('Error %s: Error of parameters. Current or Voltage expected,'
                   'other given' % value)
        elif value == ERR_KEITHLEY_INVALID_STATUS:
            msg = 'Error %s: Error of Status. Status not known' % value
        elif value == ERR_KEITHLEY_BAD_PARAMS:
            msg = ('Error %s: Error of parameters. Parameter not writable '
                   'given' % value)
        elif value == ERR_KEITHLEY_BAD_CHANNEL:
            msg = ('Error %s: Error of parameter selection. Parameter must '
                   'be between P6V, P25V and N25V' % value)
        elif value == ERR_KEITHLEY_GPIB_WRITE_READ:
            msg = ('Error %s: Error of GPIB message when writing and reading.'
                   % value)
        elif value == ERR_KEITHLEY_GPIB_READ:
            msg = ('Error %s: GPIB error when reading' % value)
        elif value == ERR_KEITHLEY_GPIB_WRITE:
            msg = ('Error %s: Error of GPIB when writing command '
                   'given.' % value)
        elif value == ERR_KEITHLEY_BAD_MODEL:
            msg = ('Error %s: Error of Keithley model. This command is not '
                   'accessible for this model' % value)
        elif value == ERR_KEITHLEY_WRONG_PARAMS_OUTPUT_ELEMENTS:
            msg = ('Error %s: Error of Keithley output elements parameter.' %
                   value)
        elif value == ERR_KEITHLEY_WRONG_PARAMS_FUNCTION:
            msg = ('Error %s: Error of Keithley function parameter. Function '
                   'must be one of the following: \n %s' % (value, var))
        elif value == ERR_KEITHLEY_WRONG_PARAMS_OUTPUT_TYPE:
            msg = ('Error %s: Error of Keithley output type parameter. '
                   'Output type must be one of the following: \n %s'
                   % (value, var))
        elif value == ERR_KEITHLEY_WRONG_PARAMS_POWERON:
            msg = ('Error %s: Error of Keithley poweron bad parameter. Must '
                   'be either "RST" or "PRES"' % value)
        elif value == ERR_KEITHLEY_WRONG_PARAMS_AUTOZERO:
            msg = ('Error %s: Error on Keithley autozero parameter. Must be '
                   'either "NORM" or "SYNC"' % value)
        elif value == ERR_KEITHLEY_WRONG_PARAMS_AMETHOD:
            msg = ('Error %s: Error on Keithley acquisition method. Parameter '
                   'must be one of the following: \n %s' % (value, var))
        elif value == ERR_KEITHLEY_WRONG_PARAMS_INPUT_SOURCE:
            msg = ('Error %s: Error on Keithley input source parameter. Value '
                   'must be one of the following: \n %s' % (value, var))
        elif value == ERR_KEITHLEY_WRONG_PARAMS_COUPLE_TRIGG:
            msg = ('Error %s: Error on Keithley trigger coupling parameter. '
                   'Must be either "ASYN" or "SSYN"' % value)
        elif value == ERR_KEITHLEY_INIT_IGNORED:
            msg = ('Error %s: Error on Keithley initiate.' % value)
        else:
            msg = 'Unknown error (%s)' % value
        return msg


class Keithley(object):
    """
    Class to control the Keithley multimeter
    """
    def __init__(self, gpib_ds_path='gpib/keithley2001/1',
                 logfile='/tmp/keithley2001.log'):
        """
        Method to initialise the object connecting to the Keithley and to
        open a log file
        :param logfile: path to the target log file
        :return:
        """

        self.gpib = PyTango.DeviceProxy(gpib_ds_path)

        # Thread required obj
        self.quit_event = threading.Event()

        # Setup logging:
        self.logfile = logfile
        self.keithleylogger = logging.getLogger('KeithleyLogger')
        self.keithleylogger.setLevel(logging.INFO)
        handler = logging.handlers.RotatingFileHandler(
                      self.logfile, maxBytes=1024*1024*10, backupCount=5,)
        formatter = logging.Formatter("%(asctime)s- %(levelname)s "
                                      "- %(message)s")
        handler.setFormatter(formatter)
        self.keithleylogger.addHandler(handler)

        self.read_value = 0

        self.list_functions = ['VOLT', 'CURR', 'VOLT:AC', 'CURR:AC', 'RES',
                            'FRES', 'TEM', 'FREQ', 'VOLT:FREQ', 'CURR:FREQ']
        self.list_output_data_types = ['ASC', 'SRE', 'DRE',]
        self.list_data_elements = ['READ', 'CHAN', 'UNIT', ]
        self.list_acq_methods = ['NORM', 'BURS', 'ASTR', 'SSTR', ]
        self.list_input_source = ['IMM', 'MAN', 'BUS', 'EXT', 'TIM', ]

        self.count_measure_limit = 10000

        try:
            if '2001' in self.identification:
                self.list_output_data_types += ['REAL,64', 'REAL,32', ]
                self.list_data_elements += ['RNUM', 'TIME', 'STAT', ]
                # self.list_acq_methods = ['NORM', 'BURS', 'ASTR', 'SSTR', ]
                self.list_input_source += ['HOLD', 'TLINK', ]
                self.count_measure_limit = 100000
        except:
            ConnectionError('Connection via GPIB error')

    # ------------------------------------------------------------------
    #   IN/OUT GPIB Commands
    # ------------------------------------------------------------------
    def _send_msg(self, cmd):

        """
        Method to send a command TO the GPIB device
        :param cmd: Command to send
        """
        try:
            self.gpib.GPIBWrite(cmd)
        except Exception:
            raise KeithleyError(ERR_KEITHLEY_GPIB_WRITE)

    def _receive_msg(self):
        """
        Method to receive a command FROM the GPIB device
        :return: The received answer
        """
        try:
            return self.gpib.GPIBRead()
        except Exception:
            raise KeithleyError(ERR_KEITHLEY_GPIB_READ)

    def _send_recieve_msg(self, cmd):
        """
        Method to receive a value FROM the GPIB device after sending a command
        :param cmd: command to execute
        :return: String with the received answer
        """
        try:
            return self.gpib.GPIBWriteRead(cmd)
        except Exception:
            raise KeithleyError(ERR_KEITHLEY_GPIB_WRITE_READ)

    # ------------------------------------------------------------------
    #   Communication Commands
    # ------------------------------------------------------------------
    def read_cmd(self):
        """
        Method to parse the raw output of the HW and return the value after
        logging it
        :return: HW answer
        """
        output = self._receive_msg()
        self.keithleylogger.info('read: ' + output)
        output = output.strip('\n')
        self.keithleylogger.info(output)
        return output

    def write_cmd(self, cmd):
        """
        Method to send the cmd to execute after self.keithleylogger it
        :param cmd: Command
        :return: None
        """
        self.keithleylogger.info('write: ' + cmd)
        self._send_msg(cmd)

    def write_read_cmd(self, cmd):
        """
        Method to send the cmd to execute after logging it and to parse the raw
        output of the HW and return the value after logging it
        :param cmd: Command
        :return: raw output
        """
        output = self._send_recieve_msg(cmd).strip('\n')
        self.keithleylogger.info('in: ' + cmd + '\t\tout: ' + output)
        return output

    # ------------------------------------------------------------------
    #   Signal oriented measurement commands
    # ------------------------------------------------------------------

    @property
    def read_postprocessed_fetch(self):
        """
        Query cmd to request post-processed reading
        :return: read value
        """
        cmd = ':fetch?'
        return self.write_read_cmd(cmd)

    @property
    def configure(self):
        """
        Query of the instrument configuration function for subsequent
        measurements.
        :return: read value
        """
        cmd = ':conf?'
        return self.write_read_cmd(cmd)

    @configure.setter
    def configure(self, function):
        """
        Cmd to configure the instrument for subsequent measurements.
        :param function: parameter to set
        :return: None
        """

        if function in self.list_functions:
            cmd = ':conf:' + str(function)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_FUNCTION,
                                self.list_functions)

    @property
    def read_postprocessed(self):
        """
        Query cmd to request post-processed reading after config
        :return: read value
        """
        cmd = ':read?'
        return self.write_read_cmd(cmd)

    def get_measure(self, function):
        """
        Query to read an instrument measure
        :param function: to be set
        :return: read value
        """
        cmd = ':meas'
        if function != '':
            cmd += ':' + str(function)
        cmd += '?'
        return self.write_read_cmd(cmd)

    # ------------------------------------------------------------------
    #   Calculate subsystems
    # ------------------------------------------------------------------

    @property
    def calculate_1_function(self):
        """
        Query programmed math format
        :return: read value
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':form?'
        return self.write_read_cmd(cmd)

    @calculate_1_function.setter
    def calculate_1_function(self, function):
        """
        Cmd to configure the function selected for calc1.
        :param function: parameter to set
        :return: None
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':form ' + str(function)
        self.write_cmd(cmd)

    # TODO implement methods to set functions parameters

    @property
    def calculate_1_enable(self):
        """
        Query if programmed math format is enabled
        :return: read value
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':state?'
        return self.write_read_cmd(cmd)

    @calculate_1_enable.setter
    def calculate_1_enable(self, value):
        """
        Cmd to enable calc1.
        :param value: parameter to set
        :return: None
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':state ' + str(value)
        self.write_cmd(cmd)

    @property
    def calculate_1_data(self):
        """
        Query if programmed math format is enabled
        :return: read value
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':data?'
        return self.write_read_cmd(cmd)

    @property
    def calculate_1_force(self):
        """
        Force to apply the math function set to the input data and returns
        the result
        :return: read value
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':imm?'
        return self.write_read_cmd(cmd)

    @calculate_1_force.setter
    def calculate_1_force(self):
        """
        Force to apply the math function set to the input data
        :return: read value
        """
        calculate = '1'
        cmd = ':calc' + calculate + ':imm'
        self.write_cmd(cmd)

    @property
    def calculate_2_function(self):
        """
        Query programmed math format
        :return: read value
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':form?'
        return self.write_read_cmd(cmd)

    @calculate_2_function.setter
    def calculate_2_function(self, function):
        """
        Cmd to configure the function selected for calc2.
        :param function: parameter to set
        :return: None
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':form ' + str(function)
        self.write_cmd(cmd)

    # TODO implement methods to set functions parameters

    @property
    def calculate_2_enable(self):
        """
        Query if programmed math format is enabled
        :return: read value
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':state?'
        return self.write_read_cmd(cmd)

    @calculate_2_enable.setter
    def calculate_2_enable(self, value):
        """
        Cmd to enable calc2.
        :param value: parameter to set
        :return: None
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':state ' + str(value)
        self.write_cmd(cmd)

    @property
    def calculate_2_data(self):
        """
        Return calc 2 data
        :return: read value
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':data?'
        return self.write_read_cmd(cmd)

    @property
    def calculate_2_force(self):
        """
        Force to apply the math function set to the input data and returns
        the result
        :return: read value
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':imm?'
        return self.write_read_cmd(cmd)

    @calculate_2_force.setter
    def calculate_2_force(self):
        """
        Force to apply the math function set to the input data
        :return: read value
        """
        calculate = '2'
        cmd = ':calc' + calculate + ':imm'
        self.write_cmd(cmd)

    # TODO implement calc3 for limits on measure that set flags when achieved

    # TODO implement calibration subsystem (see model 2001 calibration manual)

    # TODO implement display commands

    # ------------------------------------------------------------------
    #   Format subsystem
    # ------------------------------------------------------------------

    @property
    def output_data_type(self):
        """
        Query to see the selected data format for transferring readings
        over the bus.
        :return: read value
        """
        cmd = ':format:data?'
        return self.write_read_cmd(cmd)

    @output_data_type.setter
    def output_data_type(self, data_type):
        """
        Cmd to select the data format for transferring readings over the bus.
        :param data_type: parameter to set
        :return: None
        """
        if data_type in self.list_output_data_types:
            cmd = ':format:data ' + str(data_type)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_OUTPUT_TYPE,
                                self.list_output_data_types)

    @property
    def output_data_elements(self):
        """
        Query to see the selected data format for transferring readings
        over the bus.
        :return: read value
        """
        cmd = ':format:elem?'
        return self.write_read_cmd(cmd)

    @output_data_elements.setter
    def output_data_elements(self, data_elements):
        """
        Cmd to set the data format for transferring readings
        :param data_elements: parameter to set
        :return: None
        """
        cmd = ':format:elem ' + str(data_elements[0])
        for element in data_elements[1::]:
            if element in self.list_data_elements:
                cmd += ',' + str(element)
            else:
                KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_OUTPUT_ELEMENTS)
        self.write_cmd(cmd)

    @property
    def output_data_binary_order(self):
        """
        Cmd to check data bit order format on binary output modes.
        :return: None
        """
        cmd = ':format:bord?'
        return self.write_read_cmd(cmd)

    @output_data_binary_order.setter
    def output_data_binary_order(self, data_order):
        """
        Cmd to select the data bit order format on binary output modes.
        :param data_order: param to set
        :return: None
        """

        if data_order in ['NORM', 'SWAP']:
            cmd = ':format:bord ' + str(data_order)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be NORM or SWAP')

    # ------------------------------------------------------------------
    #   Output subsystem
    # ------------------------------------------------------------------

    # TODO implement output commands (polarity of output. High/low)

    # ------------------------------------------------------------------
    #   Route subsystem
    # ------------------------------------------------------------------

    # NOTE: -241,"Hardware missing"

    # def channel_is_closed(self, channels_list):
    #     """
    #     Cmd to check closed channels from list.
    #     :param channels_list: to check
    #     :return: None
    #     """
    #     cmd = ':rout:clos? (@ ' + str(channels_list[0])
    #     for channel in channels_list:
    #         cmd += ',' + str(channel)
    #     cmd += ')'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # def channel_close(self, channels_list):
    #     """
    #     Cmd to close channels from list.
    #     :param channels_list: to check
    #     :return: None
    #     """
    #     cmd = ':rout:clos (@ ' + str(channels_list[0])
    #     for channel in channels_list[1:]:
    #         cmd += ',' + str(channel)
    #     cmd += ')'
    #     self.write_cmd(cmd)
    #
    # @property
    # def channel_closed_list(self):
    #     """
    #     Cmd to check closed channels.
    #     :return: None
    #     """
    #     cmd = ':rout:clos:stat?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # def channel_is_opened(self, channels_list):
    #     """
    #     Cmd to check state of channels from list.
    #     :param channels_list: to check
    #     :return: None
    #     """
    #     cmd = ':rout:open? (@ ' + str(channels_list[0])
    #     for channel in channels_list:
    #         cmd += ',' + str(channel)
    #     cmd += ')'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # def channel_open(self, channels_list):
    #     """
    #     Cmd to open channels from list.
    #     :param channels_list: to check
    #     :return: None
    #     """
    #     cmd = ':rout:open (@ ' + str(channels_list[0])
    #     for channel in channels_list:
    #         cmd += ',' + str(channel)
    #     cmd += ')'
    #     self.write_cmd(cmd)
    #
    # def channel_open_all(self):
    #     """
    #     Cmd to open all channels.
    #     :return: None
    #     """
    #     cmd = ':rout:open:all'
    #     self.write_cmd(cmd)
    #     # TODO: implement as thread
    #
    # def channel_scan_selection(self, channels_list):
    #     """
    #     Cmd to select channels for scan.
    #     :param channels_list: to check
    #     :return: None
    #     """
    #     cmd = ':rout:scan (@ ' + str(channels_list[0])
    #     for channel in channels_list:
    #         cmd += ',' + str(channel)
    #     cmd += ')'
    #     self.write_cmd(cmd)
    #
    # def channel_external(self):
    #     pass
    #     # TODO implement if according card present.
    #
    # def channel_function(self, channels_list, channel_function):
    #     """
    #     Cmd to open channels from list.
    #     :param channels_list: to set
    #     :param channel_function: to set
    #     :return: None
    #     """
    #     cmd = ':rout:open (@ ' + str(channels_list[0])
    #     for channel in channels_list:
    #         cmd += ',' + str(channel)
    #     cmd += '), ' + str(channel_function)
    #     self.write_cmd(cmd)
    #
    # @property
    # def ratio_function(self):
    #     """
    #     Cmd to select measurement function for specific calculation
    #     :return: None
    #     """
    #     cmd = ':rout:scan:rat:func?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # @ratio_function.setter
    # def ratio_function(self, function_name):
    #     """
    #     Cmd to select measurement function for specific calculation
    #     :param function_name: to set
    #     :return: None
    #     """
    #     cmd = ':rout:scan:ratio:function' + str(function_name)
    #     self.write_cmd(cmd)
    #
    # @property
    # def delta_function(self):
    #     """
    #     Cmd to select measurement function for specific calculation
    #     :return: None
    #     """
    #     cmd = ':rout:scan:delt:func?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # @delta_function.setter
    # def delta_function(self, function_name):
    #     """
    #     Cmd to select measurement function for specific calculation
    #     :param function_name: to set
    #     :return: None
    #     """
    #     cmd = ':rout:scan:delta:function' + str(function_name)
    #     self.write_cmd(cmd)
    #
    # @property
    # def ratio_reference_channel(self):
    #     """
    #     Cmd to select reference channel for specific calculation
    #     :return: None
    #     """
    #     cmd = ':rout:scan:rat:rch?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # @ratio_reference_channel.setter
    # def ratio_reference_channel(self, channel):
    #     """
    #     Cmd to select reference channel for specific calculation
    #     :param channel: to set as reference
    #     :return: None
    #     """
    #     cmd = ':rout:scan:ratio:rch (@ ' + str(channel) + ')'
    #     self.write_cmd(cmd)
    #
    # @property
    # def delta_reference_channel(self):
    #     """
    #     Cmd to check reference channel for specific calculation
    #     :return: None
    #     """
    #     cmd = ':rout:scan:delta:rch?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # @delta_reference_channel.setter
    # def delta_reference_channel(self, channel):
    #     """
    #     Cmd to select reference channel for specific calculation
    #     :param channel: to set as reference
    #     :return: None
    #     """
    #     cmd = ':rout:scan:delta:rch (@ ' + str(channel) + ')'
    #     self.write_cmd(cmd)
    #
    # @property
    # def ratio_measure_channel(self):
    #     """
    #     Cmd to select measure channel for specific calculation
    #     :return: None
    #     """
    #     cmd = ':rout:scan:ratio:mch?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # @ratio_measure_channel.setter
    # def ratio_measure_channel(self, channel):
    #     """
    #     Cmd to select measure channel for specific calculation
    #     :param channel: to set as measure
    #     :return: None
    #     """
    #     cmd = ':rout:scan:ratio:mch (@ ' + str(channel) + ')'
    #     self.write_cmd(cmd)
    #
    # @property
    # def delta_measure_channel(self, channel):
    #     """
    #     Cmd to check measure channel for specific calculation
    #     :param channel: to set as measure
    #     :return: None
    #     """
    #     cmd = ':rout:scan:delt:mch?'
    #     return self.write_read_cmd(cmd)
    #     # TODO: implement as thread
    #
    # @delta_measure_channel.setter
    # def delta_measure_channel(self, channel):
    #     """
    #     Cmd to select measure channel for specific calculation
    #     :param channel: to set as measure
    #     :return: None
    #     """
    #     cmd = ':rout:scan:delta:mch (@ ' + str(channel) + ')'
    #     self.write_cmd(cmd)
    #
    @property
    def scan_operation(self):
        """
        Cmd to check scan operation for channel for specific calculation
        :return: None
        """
        cmd = ':rout:scan:lselect?'
        return self.write_read_cmd(cmd)

    @scan_operation.setter
    def scan_operation(self, operation):
        """
        Cmd to select operation scan mode.
        :param operation: to set as measure
        :return: None
        """
        cmd = ':rout:scan:lselect ' + str(operation)
        self.write_cmd(cmd)

    # ------------------------------------------------------------------
    #   Sense subsystem
    # ------------------------------------------------------------------

    def sense_1_save_alternate(self):
        """
        Cmd to save the current instrument setup as alternate.
        NOTE: lost on power-up or :sys:preset
        :return: None
        """
        cmd = ':sense1:alternate1:save'
        self.write_cmd(cmd)

    def sense_1_load_alternate(self):
        """
        Cmd to save the current instrument setup as alternate.
        :return: None
        """
        cmd = ':sense1:alternate1:recall'
        self.write_cmd(cmd)

    @property
    def measure_function(self):
        """
        Cmd to check the function for measuring.
        :return: None
        """
        cmd = ':sense1:function?'
        return self.write_read_cmd(cmd)

    @measure_function.setter
    def measure_function(self, function):
        """
        Cmd to select the function for measuring.
        :param function: to set as measure
        :return: None
        """
        cmd = ':sense1:function ' + '"' + str(function) + '"'
        self.write_cmd(cmd)

    @property
    def read_latest(self):
        """
        Cmd to check the latest measure.
        :return: None
        """
        cmd = ':data?'
        return self.write_read_cmd(cmd)

    # TODO: thread?
    @property
    def read_fresh(self):
        """
        Cmd to check the latest fresh measure.
        NOTE: it will wait until new data is available. It will never give
        the same value twice.
        :return: None
        """
        cmd = ':data:fresh?'
        return self.write_read_cmd(cmd)

    # Support method
    def _get_method(self, function, in_cmd):
        """
        Cmd to get the value of the function using the in_cmd
        :param function: of the cmd to get
        :param in_cmd: to send to the HW
        :return: value
        """

        cmd = function + ':' + str(in_cmd) + '?'
        return self.write_read_cmd(cmd)

    # Support method
    def _set_method(self, function, in_cmd, value):
        """
        Cmd to set the value of the function using the in_cmd
        :param value: to set
        :param function: of the command to set
        :param in_cmd: to send to the HW
        :return: None
        """
        cmd = function + ':' + str(in_cmd) + ' ' + str(value)
        self.write_cmd(cmd)

    @property
    def current_aperture(self):
        """
        Cmd to check the current dc aperture
        :return: value
        """
        setting_mode = ':curr'
        if '2001' in self.identification:
            return self._get_method(setting_mode, 'aper')
        else:
            raise KeithleyError(ERR_KEITHLEY_BAD_MODEL)

    @current_aperture.setter
    def current_aperture(self, aperture_value):
        """
        Cmd to set the current dc aperture
        :param aperture_value: to set
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aper', aperture_value)

    @property
    def voltage_aperture(self):
        """
        Cmd to check the voltage dc aperture
        :return: value
        """
        setting_mode = ':volt'
        if '2001' in self.identification:
            return self._get_method(setting_mode, 'aper')
        else:
            raise KeithleyError(ERR_KEITHLEY_BAD_MODEL)

    @voltage_aperture.setter
    def voltage_aperture(self, aperture_value):
        """
        Cmd to set the voltage dc aperture
        :param aperture_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aper', aperture_value)

    @property
    def current_auto_aperture(self):
        """
        Cmd to check the current dc aperture
        :return: value
        """
        setting_mode = ':curr'
        if '2001' in self.identification:
            return self._get_method(setting_mode, 'aper:auto')
        else:
            raise KeithleyError(ERR_KEITHLEY_BAD_MODEL)


    @current_auto_aperture.setter
    def current_auto_aperture(self, aperture_value):
        """
        Cmd to set the current dc aperture
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aper:auto', aperture_value)

    @property
    def voltage_auto_aperture(self):
        """
        Cmd to check the voltage dc aperture
        :return: value
        """
        setting_mode = ':volt'
        if '2001' in self.identification:
            return self._get_method(setting_mode, 'aper:auto')
        else:
            raise KeithleyError(ERR_KEITHLEY_BAD_MODEL)

    @voltage_auto_aperture.setter
    def voltage_auto_aperture(self, aperture_value):
        """
        Cmd to set the voltage dc aperture
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aper:auto', aperture_value)

    # TODO implement other setting_modes apertures

    @property
    def current_nplc(self):
        """
        Cmd to check the current dc nplc
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'nplc')

    @current_nplc.setter
    def current_nplc(self, nplc_value):
        """
        Cmd to set the current dc nplc
        :param nplc_value: to set
        :return: value
        """
        if 10 >= nplc_value >= 0.01:
            setting_mode = ':curr'
            self._set_method(setting_mode, 'nplc', nplc_value)
        else:
            raise AttributeError('Param must be on [0.0002, 0.2]')

    @property
    def voltage_nplc(self):
        """
        Cmd to check the voltage dc nplc
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'nplc')

    @voltage_nplc.setter
    def voltage_nplc(self, nplc_value):
        """
        Cmd to set the voltage dc nplc
        :param nplc_value: to set
        :return: value
        """
        if 10 >= nplc_value >= 0.01:
            setting_mode = ':volt'
            self._set_method(setting_mode, 'nplc', nplc_value)
        else:
            raise AttributeError('Param must be on [0.0002, 0.2]')

    @property
    def current_auto_nplc(self):
        """
        Cmd to check the current dc nplc
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'nplc:auto')

    @current_auto_nplc.setter
    def current_auto_nplc(self, nplc_value):
        """
        Cmd to set the current dc nplc
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'nplc:auto', nplc_value)

    @property
    def voltage_auto_nplc(self):
        """
        Cmd to check the voltage dc nplc
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'nplc:auto')

    @voltage_auto_nplc.setter
    def voltage_auto_nplc(self, nplc_value):
        """
        Cmd to set the voltage dc nplc
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'nplc:auto', nplc_value)

    # TODO implement other setting_modes nplc

    @property
    def coupled_voltage_ac(self):
        """
        Cmd to check if the voltage ac is coupled
        :return: value
        """
        cmd = ':volt:ac:coup?'
        return self.write_read_cmd(cmd)

    @coupled_voltage_ac.setter
    def coupled_voltage_ac(self, param):
        """
        Cmd to couple the voltage.
        When AC coupling is selected, a DC blocking capacitor is placed in
        series with the AC measurement circuit. This removes the DC component
        from the measurement. When DC coupling is selected, the blocking
        capacitor is removed. Subsequent measurements will reflect both the
        AC and DC components of the signal (AC+DC).
        :param param: to couple
        :return: None
        """
        if param in ['AC', 'DC']:
            cmd = ':volt:ac:coup ' + str(param)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be either AC or DC')

    @property
    def coupled_current_ac(self):
        """
        Cmd to check if the current ac is coupled
        :return: value
        """
        cmd = ':volt:ac:coup?'
        return self.write_read_cmd(cmd)

    @coupled_current_ac.setter
    def coupled_current_ac(self, param):
        """
        Cmd to couple the current
        :param param: to couple
        :return: None
        """
        if param in ['AC', 'DC']:
            cmd = ':curr:ac:coup ' + str(param)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be either AC or DC')

    @property
    def current_rang(self):
        """
        Cmd to check the current dc rang
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'rang')

    @current_rang.setter
    def current_rang(self, rang_value):
        """
        Cmd to set the current dc rang
        :param rang_value: to set
        :return: value
        """
        if 0.21 >= rang_value >= 0:
            setting_mode = ':curr'
            self._set_method(setting_mode, 'rang', rang_value)
        else:
            raise AttributeError('Param must be inside [0, 0.21] interval')

    @property
    def voltage_rang(self):
        """
        Cmd to check the voltage dc rang
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'rang')

    @voltage_rang.setter
    def voltage_rang(self, rang_value):
        """
        Cmd to set the voltage dc rang
        :param rang_value: to set
        :return: value
        """
        if 1100 >= rang_value >= 0:
            setting_mode = ':volt'
            self._set_method(setting_mode, 'rang', rang_value)
        else:
            raise AttributeError('Param must be inside [0, 1100] interval')

    @property
    def current_auto_rang(self):
        """
        Cmd to check the current dc rang
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'rang:auto')

    @current_auto_rang.setter
    def current_auto_rang(self, rang_value):
        """
        Cmd to set the current dc rang
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'rang:auto', rang_value)

    @property
    def voltage_auto_rang(self):
        """
        Cmd to check the voltage dc rang
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'rang:auto')

    @voltage_auto_rang.setter
    def voltage_auto_rang(self, rang_value):
        """
        Cmd to set the voltage dc rang
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'rang:auto', rang_value)

    # TODO implement other setting_modes range

    # TODO implement other auto range methods

    @property
    def voltage_reference(self):
        """
        Cmd to check the voltage dc reference value
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'ref')

    @voltage_reference.setter
    def voltage_reference(self, reference_value):
        """
        Cmd to set the voltage dc reference value
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'ref', reference_value)

    @property
    def current_reference(self):
        """
        Cmd to check the current dc reference value
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'ref')

    @current_reference.setter
    def current_reference(self, reference_value):
        """
        Cmd to set the current dc reference value
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'ref', reference_value)

    @property
    def voltage_reference_enable(self):
        """
        Cmd to check the voltage dc reference enable
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'ref:stat')

    @voltage_reference_enable.setter
    def voltage_reference_enable(self, reference_value):
        """
        Cmd to set the voltage dc reference enable
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'ref:stat', reference_value)

    @property
    def current_reference_enable(self):
        """
        Cmd to see if the current dc reference is enabled
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'ref:stat')

    @current_reference_enable.setter
    def current_reference_enable(self, reference_value):
        """
        Cmd to set the current dc reference enable flag
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'ref:stat', reference_value)

    def voltage_reference_acquire(self):
        """
        Cmd to set the voltage dc reference with the actual input
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'ref:acq', '')
        return self.voltage_reference

    def current_reference_acquire(self):
        """
        Cmd to set the current dc reference with the actual input
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'ref:acq', '')
        return self.current_reference

    @property
    def voltage_digits(self):
        """
        Cmd to check the actual voltage digits set as precision
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'dig')

    @voltage_digits.setter
    def voltage_digits(self, digits_value):
        """
        Cmd to set the voltage digits for precision
        :param digits_value: value to set
        :return: value
        """
        if 0 <= digits_value <= 7:
            setting_mode = ':volt'
            self._set_method(setting_mode, 'dig', digits_value)
        else:
            raise AttributeError('Param must be an int inside [4, 7] interval')

    @property
    def current_digits(self):
        """
        Cmd to check the actual current digits set as precision
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'dig')

    @current_digits.setter
    def current_digits(self, digits_value):
        """
        Cmd to set the current digits for precision
        :param digits_value: value to set
        :return: value
        """
        if 0 <= digits_value <= 7:
            setting_mode = ':curr'
            self._set_method(setting_mode, 'dig', digits_value)
        else:
            raise AttributeError('Param must be an int inside [4, 7] interval')
        self._set_method(setting_mode, 'dig', digits_value)

    @property
    def voltage_auto_digits(self):
        """
        Cmd to check the actual voltage digits set as precision automatically
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'dig:auto')

    @voltage_auto_digits.setter
    def voltage_auto_digits(self, digits_value):
        """
        Cmd to set the voltage digits for precision automatically
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'dig:auto', digits_value)

    @property
    def current_auto_digits(self):
        """
        Cmd to check the actual current digits set as precision automatically
        :return: value
        """
        setting_mode = ':curr'
        return eval(self._get_method(setting_mode, 'dig:auto'))

    @current_auto_digits.setter
    def current_auto_digits(self, digits_value):
        """
        Cmd to set the current digits for precision automatically
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'dig:auto', digits_value)

    # TODO: average options commands

    @property
    def current_count(self):
        """
        Cmd to check the actual count of measures for average value
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'aver:count')

    @current_count.setter
    def current_count(self, count_value):
        """
        Cmd to set the count of measures for average value
        :param count_value: to set
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aver:count', count_value)

    @property
    def voltage_count(self):
        """
        Cmd to check the actual count of measures for average value
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'aver:count')

    @voltage_count.setter
    def voltage_count(self, count_value):
        """
        Cmd to set the count of measures for average value
        :param count_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aver:count', count_value)

    @property
    def current_noise_tolerance(self):
        """
        Cmd to check the actual noise tolerance
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'aver:adv:ntol')

    @current_noise_tolerance.setter
    def current_noise_tolerance(self, percent_value):
        """
        Cmd to set the actual noise tolerance
        :param percent_value: to set
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aver:adv:ntol', percent_value)

    @property
    def voltage_noise_tolerance(self):
        """
        Cmd to check the actual noise tolerance
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'aver:adv:ntol')

    @voltage_noise_tolerance.setter
    def voltage_noise_tolerance(self, percent_value):
        """
        Cmd to set the actual noise tolerance
        :param percent_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aver:adv:ntol', percent_value)

    @property
    def current_advanced_filters(self):
        """
        Cmd to check the actual advanced filters state
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'aver:adv')

    @current_advanced_filters.setter
    def current_advanced_filters(self, state_value):
        """
        Cmd to set the actual advanced filters state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aver:adv', state_value)

    @property
    def voltage_advanced_filters(self):
        """
        Cmd to check the actual advanced filters state
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'aver:adv')

    @voltage_advanced_filters.setter
    def voltage_advanced_filters(self, state_value):
        """
        Cmd to set the actual advanced filters state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aver:adv', state_value)

    @property
    def current_average(self):
        """
        Cmd to check the average digital filter status
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'aver')

    @current_average.setter
    def current_average(self, state_value):
        """
        Cmd to set the average digital filter state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aver', state_value)

    @property
    def voltage_average(self):
        """
        Cmd to check the average digital filter state
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'aver')

    @voltage_average.setter
    def voltage_average(self, state_value):
        """
        Cmd to set the average digital filter state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aver', state_value)

    @property
    def current_auto_average(self):
        """
        Cmd to check the average digital filter status
        :return: value
        """
        setting_mode = ':curr'
        return self._get_method(setting_mode, 'aver:auto')

    @current_auto_average.setter
    def current_auto_average(self, state_value):
        """
        Cmd to set the average digital filter state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':curr'
        self._set_method(setting_mode, 'aver:auto', state_value)

    @property
    def voltage_auto_average(self):
        """
        Cmd to check the average digital filter state
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'aver:auto')

    @voltage_auto_average.setter
    def voltage_auto_average(self, state_value):
        """
        Cmd to set the average digital filter state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'aver:auto', state_value)

    @property
    def voltage_analog_filter(self):
        """
        Cmd to check the average digital filter state
        :return: value
        """
        setting_mode = ':volt'
        return self._get_method(setting_mode, 'filt')

    @voltage_analog_filter.setter
    def voltage_analog_filter(self, state_value):
        """
        Cmd to set the average digital filter state
        :param state_value: to set
        :return: value
        """
        setting_mode = ':volt'
        self._set_method(setting_mode, 'filt', state_value)

    @property
    def current_detector_function(self):
        """
        Cmd to check the measurement type for ACI and ACV
        :return: value
        """
        setting_mode = ':curr:ac'
        return self._get_method(setting_mode, 'det')

    @current_detector_function.setter
    def current_detector_function(self, function_value):
        """
        Cmd to set the measurement type for ACI and ACV
        :param function_value: to set
        :return: value
        """
        setting_mode = ':curr:ac'
        self._set_method(setting_mode, 'det', function_value)

    @property
    def voltage_detector_function(self):
        """
        Cmd to check the measurement type for ACI and ACV
        :return: value
        """
        setting_mode = ':volt:ac'
        return self._get_method(setting_mode, 'aver:auto')

    @voltage_detector_function.setter
    def voltage_detector_function(self, function_value):
        """
        Cmd to set the measurement type for ACI and ACV
        :param function_value: to set
        :return: value
        """
        setting_mode = ':volt:ac'
        self._set_method(setting_mode, 'aver:auto', function_value)

    # TODO method to display spike on window :PWINdow <n>

    @property
    def current_dci_method(self):
        """
        Cmd to check the measurement type for ACI and ACV
        :return: value
        """
        setting_mode = ':curr:dc'
        return self._get_method(setting_mode, 'meth')

    @current_dci_method.setter
    def current_dci_method(self, function_value):
        """
        Cmd to set the measurement type for ACI and ACV
        :param function_value: to set
        :return: value
        """
        setting_mode = ':curr:dc'
        self._set_method(setting_mode, 'meth', function_value)

    @property
    def input_source(self):
        """
        Cmd to get the input measure type
        :return: value
        """
        cmd = ':freq:sour?'
        return self.write_read_cmd(cmd)

    @input_source.setter
    def input_source(self, source_value):
        """
        Cmd to set the input measure type (curr or volt)
        :param source_value: to set
        :return: None
        """
        cmd = ':freq:source ' + str(source_value)
        self.write_cmd(cmd)

    @property
    def current_input_range(self):
        """
        Cmd to check the input range
        :return: value
        """
        setting_mode = ':freq:thr:curr'
        return self._get_method(setting_mode, 'rang')

    @current_input_range.setter
    def current_input_range(self, function_value):
        """
        Cmd to set the input range
        :param function_value: to set
        :return: value
        """
        if 1000 >= function_value >= 0:
            setting_mode = ':freq:thr:curr'
            self._set_method(setting_mode, 'rang', function_value)
        else:
            raise AttributeError('Param must be inside [0, 1000] interval')


    @property
    def voltage_input_range(self):
        """
        Cmd to check the input range
        :return: value
        """
        setting_mode = ':freq:thr:volt'
        return self._get_method(setting_mode, 'rang')

    @voltage_input_range.setter
    def voltage_input_range(self, function_value):
        """
        Cmd to set the input range
        :param function_value: to set
        :return: value
        """
        if 1000 >= function_value >= 0:
            setting_mode = ':freq:thr:volt'
            self._set_method(setting_mode, 'rang', function_value)
        else:
            raise AttributeError('Param must be inside [0, 1000] interval')


    @property
    def current_input_threshold(self):
        """
        Cmd to check the input threshold
        :return: value
        """
        setting_mode = ':freq:thr:curr'
        return self._get_method(setting_mode, 'lev')

    @current_input_threshold.setter
    def current_input_threshold(self, function_value):
        """
        Cmd to set the input threshold
        :param function_value: to set
        :return: value
        """
        setting_mode = ':freq:thr:curr'
        self._set_method(setting_mode, 'lev', function_value)

    @property
    def voltage_input_threshold(self):
        """
        Cmd to check the input threshold
        :return: value
        """
        setting_mode = ':freq:thr:volt'
        return self._get_method(setting_mode, 'lev')

    @voltage_input_threshold.setter
    def voltage_input_threshold(self, function_value):
        """
        Cmd to set the input threshold
        :param function_value: to set
        :return: value
        """
        setting_mode = ':freq:thr:volt'
        self._set_method(setting_mode, 'lev', function_value)

    # TODO transducer setting
    # TODO RDT standard and other factors + parameters
    # TODO SPRTD thermometers

    def digital_input_threshold(self):
        """
        Cmd to set the digital input port threshold
        :return: value
        """
        cmd = ':freq:thr:volt:ttl'
        return self.write_read_cmd(cmd)

    @property
    def digital_input_port(self):
        """
        Cmd to check the digital input port status
        :return: value
        """
        cmd = ':sens2:ttl:data?'
        return self.write_read_cmd(cmd)

    # ------------------------------------------------------------------
    #   Source subsystem
    # ------------------------------------------------------------------

    @property
    def output_port_polarity(self, line='1'):
        """
        Cmd to check the output port polarity threshold
        :param line: to check
        :return: value
        """
        cmd = ':sour:ttl' + str(line) + '?'
        return self.write_read_cmd(cmd)

    @output_port_polarity.setter
    def output_port_polarity(self, new_value,  line='1'):
        """
        Cmd to check the output port polarity threshold
        :param new_value: value to set
        :param line: to set
        :return: value
        """
        cmd = ':sour:ttl' + str(line) + ' ' + str(new_value)
        self.write_cmd(cmd)

    # ------------------------------------------------------------------
    #   Status subsystem
    # ------------------------------------------------------------------
    @property
    def measurement_event_register(self):
        """
        Cmd to Read Measurement Event Register
        :return: value
        """
        cmd = ':stat:meas?'
        return self.write_read_cmd(cmd)

    @property
    def measurement_event_register_enable(self):
        """
        Cmd to Read Measurement Event Register Enable
        :return: value
        """
        cmd = ':stat:meas:enab?'
        return self.write_read_cmd(cmd)

    @measurement_event_register_enable.setter
    def measurement_event_register_enable(self, enable_bits):
        """
        Cmd to Read Measurement Event Register Enable
        :return: value
        """
        cmd = ':stat:meas:enab ' + str(enable_bits)
        self.write_cmd(cmd)

    # TODO: Registers parsing

    @property
    def trigger_event_register(self):
        """
        Cmd to Read Trigger event register
        :return: value
        """
        cmd = ':stat:oper:trig?'
        return eval(self.write_read_cmd(cmd))

    @property
    def operation_event_register(self):
        """
        Cmd to Read Trigger event register
        :return: value
        """
        cmd = ':stat:oper?'
        return eval(self.write_read_cmd(cmd))

    # ------------------------------------------------------------------
    #   System subsystem
    # ------------------------------------------------------------------

    def preset(self):
        """
        Cmd to return the instrument to states optimized for front panel
        operation
        :return: None
        """
        cmd = ':syst:pres'
        self.write_cmd(cmd)

    @property
    def power_on_state(self):
        """
        Cmd to check default power on setup
        :return: None
        """
        cmd = ':syst:pos?'
        return self.write_read_cmd(cmd)

    @power_on_state.setter
    def power_on_state(self, power_default):
        """
        Cmd to set default power on setup
        :param power_default: value to set
        :return: None
        """
        if power_default in ['RST', 'PRES']:
            cmd = ':syst:pos' + ' ' + str(power_default)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_POWERON)

    # TODO: switch position, system version

    @property
    def error_msg(self):
        """
        Cmd to return the last error message in a FIFO register
        :return: None
        """
        cmd = ':syst:err?'
        return self.write_read_cmd(cmd)

    def error_msg_clear(self):
        """
        Cmd to clear the error message register
        :return: None
        """
        cmd = ':syst:cle'
        self.write_cmd(cmd)

    @property
    def auto_zero(self):
        """
        Cmd to check the auto zero configuration
        :return: value
        """
        cmd = ':syst:azer:type?'
        return self.write_read_cmd(cmd)

    @auto_zero.setter
    def auto_zero(self, new_value):
        """
        Cmd to set the autozero configuration
        :param new_value: value to set
        :return: value
        """
        if new_value in ['NORM', 'SYNC']:
            cmd = ':syst:azer:type ' + str(new_value)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_AUTOZERO)

    @property
    def auto_zero_enable(self):
        """
        Cmd to check the auto zero status
        :return: value
        """
        cmd = ':syst:azer:stat?'
        return self.write_read_cmd(cmd)

    @auto_zero_enable.setter
    def auto_zero_enable(self, new_value):
        """
        Cmd to enable/disable the autozero
        :param new_value: value to set
        :return: value
        """
        cmd = ':syst:azer:stat ' + str(new_value)
        self.write_cmd(cmd)

    @property
    def acquisition_method(self):
        """
        Cmd to check the acquisition method
        :return: value
        """
        cmd = ':syst:amet?'
        return self.write_read_cmd(cmd)

    @acquisition_method.setter
    def acquisition_method(self, new_acq_value):
        """
        Cmd to set a new acquisition method
        :param new_acq_value: value to set
        :return: value
        """
        if new_acq_value in self.list_acq_methods:
            cmd = ':syst:amet ' + str(new_acq_value)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_AMETHOD,
                                self.list_acq_methods )

    @property
    def line_synchronization(self):
        """
        Cmd to check the line_synchronization status
        :return: value
        """
        cmd = ':syst:lsyn:stat?'
        return self.write_read_cmd(cmd)

    @line_synchronization.setter
    def line_synchronization(self, new_sync_value):
        """
        Cmd to set a new line_synchronization value
        :param new_sync_value: value to set
        :return: value
        """
        cmd = ':syst:lsyn:stat ' + str(new_sync_value)
        self.write_cmd(cmd)

    # TODO: front panel keys pressed

    @property
    def fetch_format(self):
        """
        Cmd to check the fetch_format
        :return: value
        """
        cmd = ':syst:fetch:form?'
        return self.write_read_cmd(cmd)

    @fetch_format.setter
    def fetch_format(self, new_fetch_value):
        """
        Cmd to set a new fetch_format value
        :param new_fetch_value: value to set
        :return: value
        """
        cmd = ':syst:fetch:form ' + str(new_fetch_value)
        self.write_cmd(cmd)

    # ------------------------------------------------------------------
    #   Trace subsystem
    # ------------------------------------------------------------------

    def buffer_clear(self):
        """
        Cmd to clear the buffer
        :return: None
        """
        cmd = ':trac:cle'
        return self.write_cmd(cmd)

    @property
    def buffer_free(self):
        """
        Cmd to check available and occupied buffer bytes
        :return: value
        """
        cmd = ':trac:free?'
        return_value = self.write_read_cmd(cmd)
        return return_value

    @property
    def buffer_elem_group(self):
        """
        Cmd to check the element group that is sent with each reading
        :return: value
        """
        cmd = ':trac:egr?'
        return self.write_read_cmd(cmd)

    @buffer_elem_group.setter
    def buffer_elem_group(self, new_elem_group):
        """
        Cmd to set  the element group for full (with time stamps) or compact
        :param new_elem_group: value to set
        :return: value
        """
        cmd = ':trac:egr ' + str(new_elem_group)
        self.write_cmd(cmd)

    @property
    def buffer_size(self):
        """
        Cmd to check the points on the buffer
        :return: value
        """
        cmd = ':trac:poin?'
        return int(self.write_read_cmd(cmd))

    @buffer_size.setter
    def buffer_size(self, new_buff_size):
        """
        Cmd to set the new buffer size (max, min and default are also options)
        :param new_buff_size: value to set
        :return: value
        """
        cmd = ':trac:poin ' + str(new_buff_size)
        self.write_cmd(cmd)

    @property
    def buffer_size_auto(self):
        """
        Cmd to check if auto points is enabled
        :return: value
        """
        cmd = ':trac:poin:auto?'
        return self.write_read_cmd(cmd)

    @buffer_size_auto.setter
    def buffer_size_auto(self, new_buff_size):
        """
        Cmd to enable auto buffer size
        :param new_buff_size: value to set
        :return: value
        """
        cmd = ':trac:poin:auto ' + str(new_buff_size)
        self.write_cmd(cmd)

    @property
    def buffer_feed(self):
        """
        Cmd to check the source of readings to be stored in the buffer
        :return: value
        """
        cmd = ':trac:feed?'
        return self.write_read_cmd(cmd)

    @buffer_feed.setter
    def buffer_feed(self, new_buff_feed):
        """
        Cmd to set a new source of readings to be stored in the buffer
        :param new_buff_feed: value to set
        :return: value
        """
        cmd = ':trac:feed ' + str(new_buff_feed)
        self.write_cmd(cmd)

    @property
    def buffer_pre_trigger_perc(self):
        """
        Cmd to check the percentage of buffer given to pretrigger readings
        :return: value
        """
        cmd = ':trac:feed:pret:amo?'
        return self.write_read_cmd(cmd)

    @buffer_pre_trigger_perc.setter
    def buffer_pre_trigger_perc(self, new_buff_pre_trigger):
        """
        Cmd to set a new percentage of the buffer to be used for pre-trigger
        readings
        :param new_buff_pre_trigger: value to set
        :return: value
        """
        cmd = ':trac:feed:pret:amo ' + str(new_buff_pre_trigger)
        self.write_cmd(cmd)

    @property
    def buffer_pre_trigger(self):
        """
        Cmd to check the buffer given to pretrigger readings
        :return: value
        """
        cmd = ':trac:feed:pret:amo:read?'
        return self.write_read_cmd(cmd)

    @buffer_pre_trigger.setter
    def buffer_pre_trigger(self, new_buff_pre_trigger):
        """
        Cmd to set a new number of positions of the buffer to be used for
        pre-trigger readings
        :param new_buff_pre_trigger: value to set
        :return: value
        """
        cmd = ':trac:feed:pret:amo:read ' + str(new_buff_pre_trigger)
        self.write_cmd(cmd)

    @property
    def buffer_pre_trigger_event(self):
        """
        Cmd to check the pretrigger source
        :return: value
        """
        cmd = ':trac:feed:pret:sour?'
        return self.write_read_cmd(cmd)

    @buffer_pre_trigger_event.setter
    def buffer_pre_trigger_event(self, new_buff_pre_trigger_event):
        """
        Cmd to set the pre-trigger source
        :param new_buff_pre_trigger_event: value to set
        :return: value
        """
        cmd = ':trac:feed:pret:sour ' + str(new_buff_pre_trigger_event)
        self.write_cmd(cmd)

    @property
    def buffer_control(self):
        """
        Cmd to check the buffer control parameter
        :return: value
        """
        cmd = ':trac:feed:cont?'
        return self.write_read_cmd(cmd)

    @buffer_control.setter
    def buffer_control(self, new_buff_control):
        """
        Cmd to set the buffer control parameter
        :param new_buff_control: value to set
        :return: value
        """
        cmd = ':trac:feed:cont ' + str(new_buff_control)
        self.write_cmd(cmd)

    @property
    def buffer_data(self):
        """
        Cmd to retrieve the buffer data
        :return: value
        """
        cmd = ':trac:data?'
        return self.write_read_cmd(cmd)

    # ------------------------------------------------------------------
    #   Trigger subsystem
    # ------------------------------------------------------------------

    def initiate(self):
        """
        Cmd to take the instrument out of idle state
        :return: None
        """
        cmd = ':init'
        try:
            self.write_cmd(cmd)
        except:
            raise KeithleyError(ERR_KEITHLEY_INIT_IGNORED)

    @property
    def initiate_continuous(self):
        """
        Cmd to check if continuous initiation is enabled
        :return: value
        """
        cmd = ':init:cont?'
        return self.write_read_cmd(cmd)

    @initiate_continuous.setter
    def initiate_continuous(self, new_init_cont):
        """
        Cmd to enable or disable continuous initiation
        :param new_init_cont: value to set
        :return: value
        """
        cmd = ':init:cont ' + str(new_init_cont)
        self.write_cmd(cmd)

    def abort(self):
        """
        Cmd to abort current operation and bring the instrument to idle state
        if continuous initiate disabled or arm layer 1 if enabled
        :return: None
        """
        cmd = ':abort'
        self.write_cmd(cmd)
        clear_measurement_flag = self.measurement_event_register

    def bypass_arm_control(self):
        """
        Cmd to bypass arm control layer
        :return: None
        """
        cmd = ':arm:imm'
        self.write_cmd(cmd)

    def bypass_scan_control(self):
        """
        Cmd to bypass scan control layer
        :return: None1
        """
        cmd = ':arm:lay2:imm'
        self.write_cmd(cmd)

    def bypass_measure_control(self):
        """
        Cmd to bypass measure control layer
        :return: None
        """
        cmd = ':trig:imm'
        self.write_cmd(cmd)

    @property
    def count_arm_control(self):
        """
        Cmd to check the number of loops that will be performed
        :return: None
        """
        cmd = ':arm:coun?'
        return self.write_read_cmd(cmd)

    @count_arm_control.setter
    def count_arm_control(self, new_loop_count):
        """
        Cmd to set the number of loops that will be performed
        :param new_loop_count: value to set
        :return: None
        """
        if 0 < new_loop_count < 10000:
            cmd = ':arm:coun ' + str(new_loop_count)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int insde (0, 10000) '
                                 'interval')


    @property
    def count_scan_control(self):
        """
        Cmd to check the number of loops that will be performed
        :return: None
        """
        cmd = ':arm:lay2:coun?'
        return self.write_read_cmd(cmd)

    @count_scan_control.setter
    def count_scan_control(self, new_loop_count):
        """
        Cmd to set the number of loops that will be performed
        :param new_loop_count: value to set
        :return: None
        """
        if 0 < new_loop_count < 10000:
            cmd = ':arm:lay2:coun ' + str(new_loop_count)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside (0, 10000) '
                                 'interval')

    @property
    def count_measure_control(self):
        """
        Cmd to check the number of loops that will be performed
        :return: None
        """
        cmd = ':trig:coun?'
        return self.write_read_cmd(cmd)

    @count_measure_control.setter
    def count_measure_control(self, new_loop_count):
        """
        Cmd to set the number of loops that will be performed
        :param new_loop_count: value to set
        :return: None
        """
        if 0 < new_loop_count < self.count_measure_limit:
            cmd = ':trig:coun ' + str(new_loop_count)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside (0, %d) '
                                 'interval' % self.count_measure_limit)

    @property
    def delay_scan_control(self):
        """
        Cmd to check the delay that will be performed before executing the layer
        :return: None
        """
        cmd = ':arm:lay2:del?'
        return eval(self.write_read_cmd(cmd))

    @delay_scan_control.setter
    def delay_scan_control(self, new_delay):
        """
        Cmd to set the delay that will be performed before executing the layer
        :param new_delay: value to set
        :return: None
        """
        if 0 <= new_delay <= 999999.999:
            cmd = ':arm:lay2:del ' + str(new_delay)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be inside [0, 1M) interval')

    @property
    def delay_measure_control(self):
        """
        Cmd to check the delay that will be performed before executing the layer
        :return: None
        """
        cmd = ':trig:del?'
        return eval(self.write_read_cmd(cmd))

    @delay_measure_control.setter
    def delay_measure_control(self, new_delay):
        """
        Cmd to set the delay that will be performed before executing the layer
        :param new_delay: value to set
        :return: None
        """
        if 0 <= new_delay <= 999999.999:
            cmd = ':trig:del ' + str(new_delay)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be inside [0, 1M) interval')

    @property
    def source_arm_control(self):
        """
        Cmd to check the source of the layer
        :return: None
        """
        cmd = ':arm:sour?'
        return self.write_read_cmd(cmd)

    @source_arm_control.setter
    def source_arm_control(self, new_source):
        """
        Cmd to set the new source of the layer
        :param new_source: value to set
        :return: None
        """
        if new_source in self.list_input_source:
            cmd = ':arm:sour ' + str(new_source)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_INPUT_SOURCE,
                                self.list_input_source )

    @property
    def source_scan_control(self):
        """
        Cmd to check the source of the layer
        :return: None
        """
        cmd = ':arm:lay2:sour?'
        return self.write_read_cmd(cmd)

    @source_scan_control.setter
    def source_scan_control(self, new_source):
        """
        Cmd to set the new source of the layer
        :param new_source: value to set
        :return: None
        """
        if new_source in self.list_input_source:
            cmd = ':arm:lay2:sour ' + str(new_source)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_INPUT_SOURCE)

    @property
    def source_measure_control(self):
        """
        Cmd to check the source of the layer
        :return: None
        """
        cmd = ':trig:sour?'
        return self.write_read_cmd(cmd)

    @source_measure_control.setter
    def source_measure_control(self, new_source):
        """
        Cmd to set the new source of the layer
        :param new_source: value to set
        :return: None
        """
        if new_source in self.list_input_source:
            cmd = ':trig:sour ' + str(new_source)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_INPUT_SOURCE)

    @property
    def timer_measure_layer(self):
        """
        Cmd to check the interval for the measure layer
        :return: None
        """
        cmd = ':trig:tim?'
        return eval(self.write_read_cmd(cmd))

    @timer_measure_layer.setter
    def timer_measure_layer(self, new_timer):
        """
        Cmd to set set the interval for the measure layer
        :param new_timer: value to set
        :return: None
        """
        if 0 <= new_timer <= 999999.999:
            cmd = ':trig:tim ' + str(new_timer)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be inside [0, 1M) interval')


    @property
    def timer_scan_layer(self):
        """
        Cmd to check the interval for the scan layer
        :return: None
        """
        cmd = ':arm:lay2:tim?'
        return eval(self.write_read_cmd(cmd))

    @timer_scan_layer.setter
    def timer_scan_layer(self, new_timer):
        """
        Cmd to set set the interval for the scan layer
        :param new_timer: value to set
        :return: None
        """
        if 0 <= new_timer <= 999999.999:
            cmd = ':arm:lay2:tim ' + str(new_timer)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be inside [0, 1M) interval')

    def bypass_arm_control_signal(self):
        """
        Cmd to bypass arm control layer source signal
        :return: None
        """
        cmd = ':arm:sign'
        self.write_cmd(cmd)

    def bypass_scan_control_signal(self):
        """
        Cmd to bypass scan control layer source signal
        :return: None
        """
        cmd = ':arm:lay2:sign'
        self.write_cmd(cmd)

    def bypass_measure_control_signal(self):
        """
        Cmd to bypass measure control layer source signal
        :return: None
        """
        cmd = ':trig:sign'
        self.write_cmd(cmd)

    @property
    def couple_triggers(self):
        """
        Cmd to check if in/out triggers ar coupled
        :return: None
        """
        cmd = ':trig:tcon:prot?'
        return self.write_read_cmd(cmd)

    @couple_triggers.setter
    def couple_triggers(self, new_value):
        """
        Cmd to couple in/out triggers
        :param new_value: value to set
        :return: None
        """
        if new_value in ['ASYN', 'SSYN']:
            cmd = ':trig:tcon:prot ' + str(new_value)
            self.write_cmd(cmd)
        else:
            raise KeithleyError(ERR_KEITHLEY_WRONG_PARAMS_COUPLE_TRIGG)

    @property
    def inline_arm_control(self):
        """
        Cmd to check the inline of the layer
        :return: None
        """
        cmd = ':arm:tcon:asyn:ilin?'
        return eval(self.write_read_cmd(cmd))

    @inline_arm_control.setter
    def inline_arm_control(self, new_inline):
        """
        Cmd to set the new inline of the layer
        :param new_inline: value to set
        :return: None
        """
        if 0 < new_inline < 7:
            cmd = ':arm:tcon:asyn:ilin ' + str(new_inline)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')

    @property
    def inline_scan_control(self):
        """
        Cmd to check the inline of the layer
        :return: None
        """
        cmd = ':arm:lay2:tcon:asyn:ilin?'
        return eval(self.write_read_cmd(cmd))

    @inline_scan_control.setter
    def inline_scan_control(self, new_inline):
        """
        Cmd to set the new inline of the layer
        :param new_inline: value to set
        :return: None
        """
        if 0 < new_inline < 7:
            cmd = ':arm:lay2:tcon:asyn:ilin ' + str(new_inline)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')

    @property
    def inline_measure_control(self):
        """
        Cmd to check the inline of the layer
        :return: None
        """
        cmd = ':trig:tcon:asyn:ilin?'
        return eval(self.write_read_cmd(cmd))

    @inline_measure_control.setter
    def inline_measure_control(self, new_inline):
        """
        Cmd to set the new inline of the layer
        :param new_inline: value to set
        :return: None
        """
        if 0 < new_inline < 7:
            cmd = ':trig:tcon:asyn:ilin ' + str(new_inline)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')

    @property
    def outline_arm_control(self):
        """
        Cmd to check the outline of the layer
        :return: None
        """
        cmd = ':arm:tcon:asyn:olin?'
        return eval(self.write_read_cmd(cmd))

    @outline_arm_control.setter
    def outline_arm_control(self, new_outline):
        """
        Cmd to set the new outline of the layer
        :param new_outline: value to set
        :return: None
        """
        if 0 < new_outline < 7:
            cmd = ':arm:tcon:asyn:olin ' + str(new_outline)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')

    @property
    def outline_scan_control(self):
        """
        Cmd to check the outline of the layer
        :return: None
        """
        cmd = ':arm:lay2:tcon:asyn:olin?'
        return eval(self.write_read_cmd(cmd))

    @outline_scan_control.setter
    def outline_scan_control(self, new_outline):
        """
        Cmd to set the new outline of the layer
        :param new_outline: value to set
        :return: None
        """
        if 0 < new_outline < 7:
            cmd = ':arm:lay2:tcon:asyn:olin ' + str(new_outline)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')

    @property
    def outline_measure_control(self):
        """
        Cmd to check the outline of the layer
        :return: None
        """
        cmd = ':trig:tcon:asyn:olin?'
        return eval(self.write_read_cmd(cmd))

    @outline_measure_control.setter
    def outline_measure_control(self, new_outline):
        """
        Cmd to set the new outline of the layer
        :param new_outline: value to set
        :return: None
        """
        if 0 < new_outline < 7:
            cmd = ':trig:tcon:asyn:olin ' + str(new_outline)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')

    @property
    def semisynchronous_line(self):
        """
        Cmd to check the outline of the layer
        :return: None
        """
        cmd = ':trig:tcon:ssyn:line?'
        return eval(self.write_read_cmd(cmd))

    @semisynchronous_line.setter
    def semisynchronous_line(self, new_line):
        """
        Cmd to set the new outline of the layer
        :param new_line: value to set
        :return: None
        """
        if 0 < new_line < 7:
            cmd = ':trig:tcon:ssyn:line ' + str(new_line)
            self.write_cmd(cmd)
        else:
            raise AttributeError('Param must be an int inside [1, 6] interval')


    # ------------------------------------------------------------------
    #   Unit subsystem
    # ------------------------------------------------------------------

    # TODO: implement methods to change units (e.g V to dB)

    # ------------------------------------------------------------------
    #   IEE-488.2 Common Commands
    # ------------------------------------------------------------------
    def clear_status(self):
        """
        Clears all event registers, and Error Queue.
        :return: None
        """
        cmd = '*CLS'
        self.write_cmd(cmd)

    @property
    def event_enable_register(self):
        """
        Request the programmed value of the Standard Event Enable Register.
        :return: None
        """
        cmd = '*ESE?'
        return eval(self.write_read_cmd(cmd))

    @event_enable_register.setter
    def event_enable_register(self, nrf):
        """
        Sets the contents of the Standard Event Enable Register.
        :param nrf: value to set
        :return: None
        """
        cmd = '*ESE ' + str(nrf)
        self.write_cmd(cmd)

    @property
    def event_status_register(self):
        """
        Request the programmed value of the Standard Event Status Register and
        clears it.
        :return: None
        """
        cmd = '*ESR?'
        return eval(self.write_read_cmd(cmd))

    @property
    def identification(self):
        """
        Request the manufacturer, model number, serial number, and firmware
        revision levels of the unit.
        :return: None
        """
        cmd = '*IDN?'
        return self.write_read_cmd(cmd)

    @property
    def operation_complete(self):
        """
        Places an ASCII “1” into the output queue when all pending selected
        device operations have been completed.
        :return: None
        """
        cmd = '*OPC?'
        return eval(self.write_read_cmd(cmd))

    @operation_complete.setter
    def operation_complete(self, value):
        """
        Sets the Operation Complete bit in the Standard Event Status Register
        after all pending commands have been executed.
        :return: None
        """
        cmd = '*OPC'
        self.write_cmd(cmd)

    @property
    def option_identification(self):
        """
        Returns an ID code that indicates which memory option is installed and
        whether or not the optional scanner card is installed.
        :return: None
        """
        cmd = '*OPT?'
        return self.write_read_cmd(cmd)

    def recall(self, nrf):
        """
        Returns the Model 2001 to the setup configuration stored in the
        designated memory location.
        :param nrf: value to set
        :return: None
        """
        cmd = '*RCL ' + str(nrf)
        self.write_cmd(cmd)

    def reset(self):
        """
        Returns the Model 2001 to the *RST default conditions.
        :return: None
        """
        cmd = '*RST'
        self.write_cmd(cmd)

    def save(self, nrf):
        """
        Saves the current setup configuration to the designated memory location.
        :param nrf: value to set
        :return: None
        """
        cmd = '*SAV ' + str(nrf)
        self.write_cmd(cmd)

    @property
    def service_request_enable_register(self):
        """
        The Model 2001 returns the value of the Service Request Enable
        Register.
        :return: None
        """
        cmd = '*SRE?'
        return eval(self.write_read_cmd(cmd))

    @service_request_enable_register.setter
    def service_request_enable_register(self, nrf):
        """
        Sets the contents of the Service Request Enable Register.
        :param nrf: value to set
        :return: None
        """
        cmd = '*SRE ' + str(nrf)
        self.write_cmd(cmd)

    @property
    def status_byte_register(self):
        """
        Returns the value of the Status Byte Register.
        :return: None
        """
        cmd = '*STB?'
        return self.write_read_cmd(cmd)

    def trigger(self):
        """
        This command issues a bus trigger.
        :return: None
        """
        cmd = '*TRG'
        self.write_cmd(cmd)

    def test(self):
        """
        When this query is sent, the Model 2001 will perform a checksum test on
        ROM and return the results.
        :return: None
        """
        self.test_value = -1
        t = threading.Thread(target=self._hardware_auto_test())
        t.start()
        return self.test_value

    def _hardware_auto_test(self):
        """
        Method executed by the thread in charge of retrieving a read from the
        device asynchronously
        :return: read value
        """
        cmd = '*TST?'
        self.write_cmd(cmd)
        time.sleep(3)
        self.test_value = self.read_cmd()

    def wait(self):
        """
        Wait until all previous commands are executed.
        :return: None
        """
        cmd = '*WAI'
        self.write_cmd(cmd)

    # ------------------------------------------------------------------
    #   Core useful commands combination
    # ------------------------------------------------------------------

    @property
    def error_flag(self):
        """
        Core commands to return if an error has ocurred
        :return: Bool
        """
        flag = (eval(self.status_byte_register) & 2**2 != 0)
        return flag

    def hardware_nonblocking_read_config(self):
        """
        Core commands combination to configure the HW for non-blocking commands
        :return: None
        """
        if int(self.measurement_event_register_enable) & 2**5 != 1:
            self.measurement_event_register_enable = 2**5
        if int(self.service_request_enable_register) & 2**0 != 1:
            self.service_request_enable_register = 2**0
        clear_register = self.measurement_event_register

    @property
    def read_nonblocking(self):
        """
        Core commands combination to read a measure without blocking the HW
        :return: None
        """
        self.read_value = 0
        t = threading.Thread(target=self._read_nonblock())
        t.start()
        return self.read_value

    def _read_nonblock(self):
        """
        Method executed by the thread in charge of retrieving a read from the
        device asynchronously
        :return: read value
        """
        sleep_time = eval(self.voltage_nplc) * 0.2 / 100
        while not int(self.status_byte_register) & 2**0 != 0:
            time.sleep(sleep_time)
        self.read_value = self.read_fresh
        clear_measurement_flag = self.measurement_event_register

    def acquisition(self, scan_points):
        """
        Core commands combination to start an acquisition
        :param scan_points: number of measures to be retrieved
        :return: Measured values
        """
        self.quit_event.clear()
        self.scan_values = []
        self.initiate()
        t = threading.Thread(target=self._acquisition_thread,
                             args=[self.quit_event])
        t.start()
        for point in range(scan_points):
            self.scan_values += [self.acquisitions_queue.get()]
        self.quit_event.set()
        self.abort()
        # print(self.acquisitions_queue.get())
        return self.scan_values

    def _acquisition_thread(self, quit_event):
        """
        Method executed by the thread in charge of retrieving readings from the
        device asynchronously
        :return: read value
        """
        sleep_time = eval(self.voltage_nplc) * 0.2 / 100
        while not quit_event.isSet():
            while not int(self.status_byte_register) & 2**0 != 0:
                time.sleep(sleep_time)
            self.acquisitions_queue.put(self.read_nonblocking)
            clear_measurement_flag = self.measurement_event_register

    def multiple_last_reads(self, number_of_reads):
        """
        Core commands combination to read multiple values on the same command
        :param number_of_reads: number of measures to be retrieved
        :return: Measured values
       """

        cmd = ':data?; ' * number_of_reads
        return self.write_read_cmd(cmd).split(';')

    def multiple_fresh_reads(self, number_of_reads):
        """
        Core commands combination to start an acquisition
        :param number_of_reads: number of measures to be retrieved
        :return: Measured values
       """

        cmd = ':data:fresh?; ' * number_of_reads
        return self.write_read_cmd(cmd).split(';')

